"""
PaCMAP-GPU: GPU-Accelerated Pairwise Controlled Manifold Approximation

A high-performance GPU implementation of PaCMAP using PyTorch and CUDA graphs
for maximum efficiency on large-scale dimensionality reduction tasks.

Main Components:
    - PaCMAPGPU: The main dimensionality reduction class
    - check_gpu_availability: Check if CUDA is available
    - get_gpu_info: Get detailed GPU device information  
    - clear_gpu_memory: Clear GPU memory cache

Example Usage:
    >>> from pacmap_gpu import PaCMAPGPU, check_gpu_availability
    >>> import numpy as np
    >>> 
    >>> # Check if GPU is available
    >>> if check_gpu_availability():
    ...     # Create and fit the model
    ...     reducer = PaCMAPGPU(n_components=2, verbose=True)
    ...     X_embedded = reducer.fit_transform(X)
    ... else:
    ...     print("CUDA not available - check GPU installation")
"""

from .pacmap_gpu import PaCMAPGPU
from .utils import (
    check_gpu_availability,
    get_gpu_info,
    clear_gpu_memory,
    verify_installation,
    print_installation_report
)

__version__ = "1.0.2"
__author__ = "Antonio Roye-Azar"
__email__ = "antonio@varonova.ca"

__all__ = [
    "PaCMAPGPU",
    "check_gpu_availability", 
    "get_gpu_info",
    "clear_gpu_memory",
    "verify_installation",
    "print_installation_report"
]

# Verify GPU availability on import
_gpu_available = check_gpu_availability()
if not _gpu_available:
    import warnings
    warnings.warn(
        "CUDA is not available. PaCMAP-GPU requires a CUDA-capable GPU and "
        "properly installed CUDA toolkit. Please check your installation.",
        UserWarning
    ) 