"""
PaCMAP-GPU: GPU-Accelerated Pairwise Controlled Manifold Approximation

This module provides a high-performance GPU implementation of PaCMAP using PyTorch
and CUDA graphs for maximum efficiency. The implementation includes both standard
PaCMAP and LocalMAP variants with highly optimized CUDA kernel execution.

Key Features:
    - CUDA Graph optimization for maximum GPU performance
    - Three-phase training with specialized optimization steps
    - Dynamic resampling for LocalMAP variant
    - Memory-efficient batch processing
    - Support for large-scale datasets (10K+ samples)

Architecture:
    - Uses PyTorch tensors and CUDA for all computations
    - Employs RAPIDS cuML for k-NN search and PCA initialization
    - Implements static CUDA graphs for minimal kernel launch overhead
    - Dynamic resampling in Phase 3 for LocalMAP behavior
"""

import torch
import numpy as np
from cuml.neighbors import NearestNeighbors as cuNearestNeighbors
from cuml.decomposition import PCA as cuPCA
from typing import Optional, Union, Tuple
import warnings


class PaCMAPGPU:
    """
    GPU-accelerated Pairwise Controlled Manifold Approximation and Projection.
    
    A high-performance implementation of PaCMAP using CUDA graphs and optimized
    GPU kernels for large-scale dimensionality reduction.
    
    Parameters:
        n_components (int): Number of dimensions in the embedding space
        n_neighbors (int): Number of nearest neighbors to consider
        MN_ratio (float): Ratio for mid-near pairs sampling
        FP_ratio (float): Ratio for further pairs sampling
        num_iters (int): Total number of optimization iterations
        batch_size (int): Batch size for stochastic optimization
        init_method (str): Initialization method ('pca' or 'random')
        optimizer (str): Optimizer type ('adam' or 'adagrad')
        lr (float): Learning rate for optimization
        device (str): CUDA device identifier
        low_dist_thres (float): Distance threshold for LocalMAP resampling
        random_state (int): Random seed for reproducibility
        verbose (bool): Whether to print progress information
    """
    
    def __init__(
        self,
        n_components: int = 2,
        n_neighbors: int = 10,
        MN_ratio: float = 0.5,
        FP_ratio: float = 2.0,
        num_iters: int = 450,
        batch_size: int = 8192,
        init_method: str = "pca",
        optimizer: str = "adam",
        lr: float = 0.01,
        device: str = "cuda",
        low_dist_thres: float = 1.0,
        random_state: Optional[int] = None,
        verbose: bool = False
    ):
        # Parameter validation
        if not isinstance(n_components, int) or n_components <= 0:
            raise ValueError("n_components must be a positive integer")
        if not isinstance(n_neighbors, int) or n_neighbors <= 0:
            raise ValueError("n_neighbors must be a positive integer")
        if not isinstance(MN_ratio, (int, float)) or MN_ratio < 0:
            raise ValueError("MN_ratio must be non-negative")
        if not isinstance(FP_ratio, (int, float)) or FP_ratio < 0:
            raise ValueError("FP_ratio must be non-negative")
        if not isinstance(num_iters, int) or num_iters < 0:
            raise ValueError("num_iters must be non-negative")
        if not isinstance(batch_size, int) or batch_size <= 0:
            raise ValueError("batch_size must be a positive integer")
        if not isinstance(lr, (int, float)) or lr <= 0:
            raise ValueError("lr must be a positive number")
        if not isinstance(low_dist_thres, (int, float)) or low_dist_thres <= 0:
            raise ValueError("low_dist_thres must be a positive number")
        
        self.n_components = n_components
        self.n_neighbors = n_neighbors
        self.MN_ratio = MN_ratio
        self.FP_ratio = FP_ratio
        self.num_iters = num_iters
        self.batch_size = batch_size
        self.init_method = init_method
        self.optimizer_name = optimizer
        self.lr = lr
        self.device = device
        self.low_dist_thres = low_dist_thres
        self.random_state = random_state
        self.verbose = verbose
        
        # Internal state
        self.embedding_ = None
        self.is_fitted = False
        
        # Set random seeds for reproducibility
        if random_state is not None:
            torch.manual_seed(random_state)
            torch.cuda.manual_seed(random_state)
            np.random.seed(random_state)
    
    def _build_neighbor_graph(self, X: np.ndarray) -> torch.Tensor:
        """
        Build k-nearest neighbor graph using RAPIDS cuML.
        
        For high-dimensional data (>128 features), applies PCA preprocessing
        to improve k-NN search efficiency while maintaining accuracy.
        
        Args:
            X: Input data array of shape (n_samples, n_features)
            
        Returns:
            Tensor of neighbor indices of shape (n_samples, 3*n_neighbors)
        """
        if self.verbose:
            print(f"Building neighbor graph for {X.shape[0]} samples...")
            
        # Apply PCA for high-dimensional data to improve k-NN efficiency
        if X.shape[1] > 128:
            if self.verbose:
                print(f"Applying PCA dimensionality reduction: {X.shape[1]} -> 128")
            pca_knn = cuPCA(n_components=128)
            X_for_knn = pca_knn.fit_transform(X)
        else:
            X_for_knn = X
        
        # Compute k-NN graph with 3x neighbors for pair sampling
        knn = cuNearestNeighbors(n_neighbors=3 * self.n_neighbors)
        knn.fit(X_for_knn)
        _, indices_cupy = knn.kneighbors(X_for_knn)
        
        return torch.as_tensor(indices_cupy, device=self.device).long()
    
    def _prepare_pair_indices(self, indices: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """
        Prepare indices for nearest neighbor (N), mid-near (MN), and further (FP) pairs.
        
        Creates index tensors for the three types of pairs used in PaCMAP:
        - N pairs: Immediate neighbors for local structure
        - MN pairs: Mid-distance pairs for intermediate structure  
        - FP pairs: Random pairs for global structure and repulsion
        
        Args:
            indices: Neighbor indices tensor of shape (n_samples, 3*n_neighbors)
            
        Returns:
            Tuple of (i_N, j_N, i_MN, j_MN, i_FP, j_FP) index tensors
        """
        n_samples = indices.shape[0]
        all_indices = torch.arange(n_samples, device=self.device)
        
        # Nearest neighbor pairs (first n_neighbors)
        i_N = all_indices.repeat_interleave(self.n_neighbors)
        j_N = indices[:, 1:self.n_neighbors + 1].flatten()
        
        # Mid-near pairs (sampled from remaining neighbors)
        n_mn_pairs = int(self.n_neighbors * self.MN_ratio)
        i_MN = all_indices.repeat_interleave(n_mn_pairs)
        mn_pool = indices[:, self.n_neighbors + 1: 3 * self.n_neighbors]
        rand_mn_indices = torch.randint(0, mn_pool.shape[1], (n_samples, n_mn_pairs), device=self.device)
        j_MN = torch.gather(mn_pool, 1, rand_mn_indices).flatten()
        
        # Further pairs (random sampling)
        n_fp_pairs = int(self.n_neighbors * self.FP_ratio)
        i_FP = all_indices.repeat_interleave(n_fp_pairs)
        j_FP = torch.randint(0, n_samples, (n_samples * n_fp_pairs,), device=self.device)
        
        return i_N, j_N, i_MN, j_MN, i_FP, j_FP
    
    def _initialize_embedding(self, X: np.ndarray) -> torch.Tensor:
        """
        Initialize the low-dimensional embedding.
        
        Uses PCA initialization for better convergence, falling back to
        random initialization for low-dimensional inputs.
        
        Args:
            X: Input data array
            
        Returns:
            Initialized embedding tensor requiring gradients
        """
        if self.verbose:
            print(f"Initializing embedding using {self.init_method} method...")
            
        if self.init_method == "pca" and X.shape[1] > self.n_components:
            pca = cuPCA(n_components=self.n_components)
            Y_cupy = pca.fit_transform(X)
            Y = torch.as_tensor(Y_cupy, device=self.device).float()
        else:
            if X.shape[1] <= self.n_components:
                warnings.warn("Input dimensionality <= output dimensionality, using random initialization")
            Y = torch.randn(X.shape[0], self.n_components, device=self.device) * 0.01
        
        Y.requires_grad = True
        return Y
    
    def _create_optimizer(self, Y: torch.Tensor) -> torch.optim.Optimizer:
        """Create optimizer for embedding parameters."""
        if self.optimizer_name.lower() == 'adam':
            return torch.optim.Adam([Y], lr=self.lr, capturable=True)
        else:
            # capturable is not supported by Adagrad in all PyTorch versions
            return torch.optim.Adagrad([Y], lr=self.lr)
    
    def _precompute_random_indices(self) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Precompute all random indices for batch sampling.
        
        Pre-generating random indices improves CUDA graph compatibility
        and reduces runtime overhead during optimization.
        
        Returns:
            Tuple of precomputed random index tensors for N, MN, and FP pairs
        """
        n_n_pairs = len(torch.arange(self.n_samples).repeat_interleave(self.n_neighbors))
        n_mn_pairs = len(torch.arange(self.n_samples).repeat_interleave(int(self.n_neighbors * self.MN_ratio)))  
        n_fp_pairs = len(torch.arange(self.n_samples).repeat_interleave(int(self.n_neighbors * self.FP_ratio)))
        
        rand_indices_N = torch.randint(0, n_n_pairs, (self.num_iters, self.batch_size), device=self.device)
        rand_indices_MN = torch.randint(0, n_mn_pairs, (self.num_iters, self.batch_size), device=self.device)
        rand_indices_FP = torch.randint(0, n_fp_pairs, (self.num_iters, self.batch_size), device=self.device)
        
        return rand_indices_N, rand_indices_MN, rand_indices_FP
    
    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Fit the model and transform the data.
        
        Args:
            X: Input data of shape (n_samples, n_features)
            
        Returns:
            Transformed data of shape (n_samples, n_components)
        """
        # Input validation
        if not isinstance(X, np.ndarray):
            raise ValueError("Input X must be a numpy array")
        if X.ndim != 2:
            raise ValueError("Input X must be a 2D array (n_samples, n_features)")
        if X.shape[0] == 0 or X.shape[1] == 0:
            raise ValueError("Input X must have non-zero shape in both dimensions")
        X = X.astype(np.float32)
        self.n_samples = X.shape[0]
        
        if self.verbose:
            print(f"Starting PaCMAP-GPU with {X.shape[0]} samples, {X.shape[1]} features")
        
        # Step 1: Build neighbor graph
        indices = self._build_neighbor_graph(X)
        
        # Step 2: Prepare pair indices
        i_N, j_N, i_MN, j_MN, i_FP, j_FP = self._prepare_pair_indices(indices)
        
        # Step 3: Initialize embedding
        Y = self._initialize_embedding(X)
        optimizer = self._create_optimizer(Y)
        
        # Step 4: Precompute random indices
        rand_indices_N, rand_indices_MN, rand_indices_FP = self._precompute_random_indices()
        
        # Step 5: Run optimization
        self._optimize_embedding(Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP, 
                               rand_indices_N, rand_indices_MN, rand_indices_FP)
        
        # Store final embedding
        self.embedding_ = Y.detach().cpu().numpy()
        self.is_fitted = True
        
        if self.verbose:
            print(f"Optimization completed. Final embedding shape: {self.embedding_.shape}")
        
        return self.embedding_
    
    def _optimize_embedding(self, Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                           rand_indices_N, rand_indices_MN, rand_indices_FP):
        """
        Main optimization loop using CUDA graphs for maximum performance.
        
        Implements three-phase training:
        - Phase 1: High MN weight for intermediate structure
        - Phase 2: High NN weight for local structure  
        - Phase 3: LocalMAP behavior with dynamic resampling
        """
        if self.verbose:
            print("Creating CUDA optimization graphs...")
        
        # Define phase boundaries
        phase1_end = int(self.num_iters * (100/450))
        phase2_end = int(self.num_iters * (200/450))
        
        # Create static index tensors for CUDA graph compatibility
        static_indices_N = torch.empty(self.batch_size, dtype=torch.long, device=self.device)
        static_indices_MN = torch.empty(self.batch_size, dtype=torch.long, device=self.device)
        static_indices_FP = torch.empty(self.batch_size, dtype=torch.long, device=self.device)
        
        # Create CUDA graphs for each phase
        g_phase1 = self._create_phase1_graph(Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                                           static_indices_N, static_indices_MN, static_indices_FP)
        
        g_phase2 = self._create_phase2_graph(Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                                           static_indices_N, static_indices_MN, static_indices_FP)
        
        g_phase3 = self._create_phase3_graph(Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                                           static_indices_N, static_indices_MN, static_indices_FP)
        
        if self.verbose:
            print("Running optimization...")
        
        # Main optimization loop
        for iter_num in range(self.num_iters):
            # Update static indices for current iteration
            static_indices_N.copy_(rand_indices_N[iter_num], non_blocking=True)
            static_indices_MN.copy_(rand_indices_MN[iter_num], non_blocking=True)
            static_indices_FP.copy_(rand_indices_FP[iter_num], non_blocking=True)
            
            # Execute appropriate phase
            if iter_num < phase1_end:
                g_phase1.replay()
            elif iter_num < phase2_end:
                g_phase2.replay()
            else:
                # Phase 3: LocalMAP with periodic resampling
                if (iter_num - phase2_end) > 0 and (iter_num - phase2_end) % 10 == 0:
                    self._update_fp_pairs_localmap(i_FP, j_FP, Y)
                g_phase3.replay()
            
            if self.verbose and (iter_num + 1) % 100 == 0:
                print(f"Iteration {iter_num + 1}/{self.num_iters}")
        
        torch.cuda.synchronize()
    
    def _create_phase1_graph(self, Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                           static_indices_N, static_indices_MN, static_indices_FP):
        """Create CUDA graph for Phase 1: High MN weight for global structure."""
        g = torch.cuda.CUDAGraph()
        with torch.cuda.graph(g):
            self._optimization_step(Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                                  2.0, 3.0, 1.0,
                                  static_indices_N, static_indices_MN, static_indices_FP)
        return g
    
    def _create_phase2_graph(self, Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                           static_indices_N, static_indices_MN, static_indices_FP):
        """Create CUDA graph for Phase 2: High NN weight for local structure."""
        g = torch.cuda.CUDAGraph()
        with torch.cuda.graph(g):
            self._optimization_step(Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                                  3.0, 1.0, 1.0,
                                  static_indices_N, static_indices_MN, static_indices_FP)
        return g
    
    def _create_phase3_graph(self, Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                           static_indices_N, static_indices_MN, static_indices_FP):
        """Create CUDA graph for Phase 3: LocalMAP behavior."""
        g = torch.cuda.CUDAGraph()
        with torch.cuda.graph(g):
            self._optimization_step_localmap(Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                                           1.0, 1.0,
                                           static_indices_N, static_indices_MN, static_indices_FP)
        return g
    
    def _optimization_step(self, Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                          w_N, w_MN, w_FP, sample_indices_N, sample_indices_MN, sample_indices_FP):
        """
        Standard PaCMAP optimization step for Phases 1-2.
        
        Implements the three-force model:
        - Attractive forces for nearest neighbors and mid-near pairs
        - Repulsive forces for further pairs
        """
        optimizer.zero_grad(set_to_none=True)
        
        total_loss = 0
        
        # Attractive force for nearest neighbors
        d_N_sq = torch.sum((Y[i_N[sample_indices_N]] - Y[j_N[sample_indices_N]]) ** 2, dim=1)
        loss_N = w_N * (d_N_sq / (10.0 + d_N_sq))
        total_loss += loss_N.mean()
        
        # Attractive force for mid-near pairs
        if w_MN > 0:
            d_MN_sq = torch.sum((Y[i_MN[sample_indices_MN]] - Y[j_MN[sample_indices_MN]]) ** 2, dim=1)
            loss_MN = w_MN * (d_MN_sq / (10000.0 + d_MN_sq))
            total_loss += loss_MN.mean()
        
        # Repulsive force for further pairs
        d_FP_sq = torch.sum((Y[i_FP[sample_indices_FP]] - Y[j_FP[sample_indices_FP]]) ** 2, dim=1)
        loss_FP = w_FP * (1.0 / (1.0 + d_FP_sq))
        total_loss += loss_FP.mean()
        
        if isinstance(total_loss, torch.Tensor):
            total_loss.backward()
            optimizer.step()

    def _optimization_step_localmap(self, Y, optimizer, i_N, j_N, i_MN, j_MN, i_FP, j_FP,
                                   w_N, w_FP, sample_indices_N, sample_indices_MN, sample_indices_FP):
        """
        LocalMAP optimization step for Phase 3.
        
        Uses modified attractive forces with distance-based weighting
        and eliminates mid-near pair contributions.
        """
        optimizer.zero_grad(set_to_none=True)
        
        total_loss = 0
        
        # Modified attractive force for nearest neighbors (LocalMAP style)
        y_ij_N = Y[i_N[sample_indices_N]] - Y[j_N[sample_indices_N]]
        d_N_sq = torch.sum(y_ij_N ** 2, dim=1)
        d_N = torch.sqrt(d_N_sq + 1e-9)  # Numerical stability
        
        # Distance-weighted attractive force
        weight = w_N * (self.low_dist_thres / 2.0) / d_N
        loss_N = (weight * d_N_sq).mean()
        total_loss += loss_N

        # Repulsive force for further pairs (no MN pairs in Phase 3)
        d_FP_sq = torch.sum((Y[i_FP[sample_indices_FP]] - Y[j_FP[sample_indices_FP]]) ** 2, dim=1)
        loss_FP = w_FP * (1.0 / (1.0 + d_FP_sq))
        total_loss += loss_FP.mean()

        if isinstance(total_loss, torch.Tensor):
            total_loss.backward()
            optimizer.step()

    @torch.no_grad()
    def _update_fp_pairs_localmap(self, i_FP, j_FP, Y, tries=10, chunk_size=8192**2):
        """
        Update further pairs for LocalMAP Phase 3.
        
        Resamples FP pairs to include only those close in the embedding space,
        implementing the LocalMAP objective of emphasizing local neighborhoods.
        
        Args:
            i_FP, j_FP: Further pair index tensors
            Y: Current embedding tensor
            tries: Maximum resampling attempts per point
            chunk_size: Processing chunk size for memory efficiency
        """
        n_samples = Y.shape[0]
        num_fp_pairs = len(i_FP)
        
        for chunk_start in range(0, num_fp_pairs, chunk_size):
            chunk_end = min(chunk_start + chunk_size, num_fp_pairs)
            chunk_slice = slice(chunk_start, chunk_end)
            
            i_fp_chunk = i_FP[chunk_slice]
            needs_resampling = torch.ones(len(i_fp_chunk), dtype=torch.bool, device=Y.device)
            
            for _ in range(tries):
                if not needs_resampling.any():
                    break

                num_needed = needs_resampling.sum()
                anchors_needing_resample = i_fp_chunk[needs_resampling]
                
                # Generate random candidates
                candidates = torch.randint(0, n_samples, (num_needed,), device=Y.device)
                
                # Check distance constraint
                dist_sq = torch.sum((Y[anchors_needing_resample] - Y[candidates])**2, dim=1)
                is_valid = dist_sq < self.low_dist_thres**2
                
                # Update j_FP where valid candidates were found
                indices_in_chunk = needs_resampling.nonzero(as_tuple=True)[0]
                j_FP[chunk_start + indices_in_chunk[is_valid]] = candidates[is_valid]
                
                # Update resampling mask
                needs_resampling[indices_in_chunk] = ~is_valid


def check_gpu_availability() -> bool:
    """Check if CUDA-capable GPU is available."""
    return torch.cuda.is_available()


def get_gpu_info() -> dict:
    """Get information about available GPU devices."""
    if not torch.cuda.is_available():
        return {"available": False, "devices": []}
    
    devices = []
    for i in range(torch.cuda.device_count()):
        props = torch.cuda.get_device_properties(i)
        devices.append({
            "id": i,
            "name": props.name,
            "memory_gb": props.total_memory / 1e9,
            "capability": f"{props.major}.{props.minor}"
        })
    
    return {
        "available": True,
        "current_device": torch.cuda.current_device(),
        "devices": devices
    }


def clear_gpu_memory():
    """Clear GPU memory cache."""
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.synchronize()