"""
Utility functions for PaCMAP-GPU.

This module provides utility functions for GPU checking, memory management,
and installation verification.
"""

import sys
import importlib
import warnings
from typing import Dict, List, Optional, Any
import torch
import numpy as np


def check_gpu_availability() -> bool:
    """
    Check if CUDA-capable GPU is available.
    
    Returns:
        True if CUDA is available and working, False otherwise
    """
    try:
        return torch.cuda.is_available() and torch.cuda.device_count() > 0
    except Exception:
        return False


def get_gpu_info() -> Dict[str, Any]:
    """
    Get comprehensive information about available GPU devices.
    
    Returns:
        Dictionary containing GPU information including device details,
        memory info, and capabilities
    """
    if not torch.cuda.is_available():
        return {"available": False, "devices": [], "error": "CUDA not available"}
    
    try:
        devices = []
        for i in range(torch.cuda.device_count()):
            props = torch.cuda.get_device_properties(i)
            memory_info = torch.cuda.mem_get_info(i) if torch.cuda.is_available() else (0, 0)

            # Robust multiprocessor count
            mp_count = getattr(props, "multiprocessor_count", None)
            if mp_count is None:
                mp_count = getattr(props, "multi_processor_count", None)

            devices.append({
                "id": i,
                "name": props.name,
                "total_memory_gb": props.total_memory / 1e9,
                "available_memory_gb": memory_info[0] / 1e9,
                "used_memory_gb": (props.total_memory - memory_info[0]) / 1e9,
                "capability": f"{props.major}.{props.minor}",
                "multiprocessor_count": mp_count,
                "max_threads_per_multiprocessor": getattr(props, "max_threads_per_multiprocessor", None),
                "max_threads_per_block": getattr(props, "max_threads_per_block", None),
            })
        
        return {
            "available": True,
            "current_device": torch.cuda.current_device(),
            "device_count": torch.cuda.device_count(),
            "devices": devices,
            "cuda_version": torch.version.cuda,
            "pytorch_version": torch.__version__
        }
    except Exception as e:
        return {
            "available": False, 
            "devices": [], 
            "error": f"Error getting GPU info: {str(e)}"
        }


def clear_gpu_memory(device_id: Optional[int] = None):
    """
    Clear GPU memory cache.
    
    Args:
        device_id: Specific GPU device ID to clear, None for all devices
    """
    if not torch.cuda.is_available():
        warnings.warn("CUDA not available, cannot clear GPU memory")
        return
        
    try:
        if device_id is not None:
            with torch.cuda.device(device_id):
                torch.cuda.empty_cache()
                torch.cuda.synchronize()
        else:
            # Clear all devices
            for i in range(torch.cuda.device_count()):
                with torch.cuda.device(i):
                    torch.cuda.empty_cache()
                    torch.cuda.synchronize()
    except Exception as e:
        warnings.warn(f"Failed to clear GPU memory: {str(e)}")


def check_rapids_installation() -> Dict[str, Any]:
    """
    Check if RAPIDS libraries are properly installed.
    
    Returns:
        Dictionary with installation status of RAPIDS components
    """
    rapids_components = {
        'cupy': 'cupy',
        'cuml': 'cuml', 
        'cudf': 'cudf',
        'faiss-gpu': 'faiss'
    }
    
    results = {}
    for name, import_name in rapids_components.items():
        try:
            module = importlib.import_module(import_name)
            version = getattr(module, '__version__', 'unknown')
            results[name] = {"installed": True, "version": version}
        except ImportError as e:
            results[name] = {"installed": False, "error": str(e)}
    
    return results


def verify_installation() -> Dict[str, Any]:
    """
    Comprehensive installation verification for PaCMAP-GPU.
    
    Returns:
        Dictionary with detailed installation status
    """
    results = {
        "python_version": sys.version,
        "python_compatible": sys.version_info >= (3, 10),
        "gpu_info": get_gpu_info(),
        "rapids_info": check_rapids_installation(),
        "core_dependencies": {},
        "overall_status": "unknown"
    }
    
    # Check core dependencies
    core_deps = ['torch', 'numpy', 'scipy', 'sklearn', 'pandas', 'matplotlib']
    for dep in core_deps:
        try:
            if dep == 'sklearn':
                module = importlib.import_module('sklearn')
            else:
                module = importlib.import_module(dep)
            version = getattr(module, '__version__', 'unknown')
            results["core_dependencies"][dep] = {"installed": True, "version": version}
        except ImportError as e:
            results["core_dependencies"][dep] = {"installed": False, "error": str(e)}
    
    # Determine overall status
    gpu_available = results["gpu_info"]["available"]
    python_ok = results["python_compatible"]
    core_deps_ok = all(info["installed"] for info in results["core_dependencies"].values())
    rapids_available = all(info["installed"] for info in results["rapids_info"].values())
    
    if python_ok and core_deps_ok and gpu_available and rapids_available:
        results["overall_status"] = "excellent"
    elif python_ok and core_deps_ok and gpu_available:
        results["overall_status"] = "good"  # Core deps + GPU available, some RAPIDS components missing
    elif python_ok and core_deps_ok:
        results["overall_status"] = "poor"  # Missing GPU support - package won't work
    else:
        results["overall_status"] = "poor"  # Major issues
    
    return results


def print_installation_report():
    """Print a detailed installation verification report."""
    print("PaCMAP-GPU Installation Verification Report")
    print("=" * 50)
    
    info = verify_installation()
    
    # Python version
    python_status = "✓" if info["python_compatible"] else "✗"
    print(f"\n{python_status} Python Version: {info['python_version'].split()[0]}")
    if not info["python_compatible"]:
        print("   ⚠ Python 3.10+ required")
    
    # GPU status
    gpu_info = info["gpu_info"]
    if gpu_info["available"]:
        print(f"\n✓ GPU Available: {gpu_info['device_count']} device(s)")
        for device in gpu_info["devices"]:
            memory_pct = (device["used_memory_gb"] / device["total_memory_gb"]) * 100
            print(f"   Device {device['id']}: {device['name']}")
            print(f"     Memory: {device['total_memory_gb']:.1f}GB "
                  f"({memory_pct:.1f}% used)")
            print(f"     Capability: {device['capability']}")
            if device.get("multiprocessor_count") is not None:
                print(f"     SM count: {device['multiprocessor_count']}")
    else:
        print("\n✗ GPU Not Available")
        if "error" in gpu_info:
            print(f"   Error: {gpu_info['error']}")
    
    # Core dependencies
    print("\n📦 Core Dependencies:")
    for dep, info_dep in info["core_dependencies"].items():
        status = "✓" if info_dep["installed"] else "✗"
        version = info_dep.get("version", "unknown")
        print(f"   {status} {dep}: {version}")
    
    # RAPIDS components
    print("\n🚀 RAPIDS Components:")
    for comp, info_comp in info["rapids_info"].items():
        status = "✓" if info_comp["installed"] else "✗"
        if info_comp["installed"]:
            version = info_comp.get("version", "unknown")
            print(f"   {status} {comp}: {version}")
        else:
            print(f"   {status} {comp}: Not installed")
    
    # Overall assessment
    status_icons = {
        "excellent": "🎉",
        "good": "👍", 
        "basic": "⚠️",
        "poor": "❌"
    }
    status_messages = {
        "excellent": "Perfect! All GPU dependencies available and working",
        "good": "Good! GPU available, some RAPIDS components may be missing",
        "poor": "Critical issues: GPU support required but not available"
    }
    
    overall = info["overall_status"]
    icon = status_icons.get(overall, "❓")
    message = status_messages.get(overall, "Unknown status")
    
    print(f"\n{icon} Overall Status: {message}")
    
    if overall != "excellent":
        print("\n💡 Recommendations:")
        if not info["python_compatible"]:
            print("   - Upgrade to Python 3.10+")
        if not gpu_info["available"]:
            print("   - Install CUDA toolkit and drivers")
            print("   - Verify GPU is CUDA-capable")
        if not any(info["rapids_info"][comp]["installed"] for comp in info["rapids_info"]):
            print("   - Install RAPIDS: conda install -c rapidsai rapids")
        missing_core = [dep for dep, info_dep in info["core_dependencies"].items() 
                       if not info_dep["installed"]]
        if missing_core:
            print(f"   - Install missing dependencies: pip install {' '.join(missing_core)}")


def get_system_info() -> Dict[str, Any]:
    """
    Get comprehensive system information for debugging.
    
    Returns:
        Dictionary with system information
    """
    import platform
    import os
    
    info = {
        "platform": {
            "system": platform.system(),
            "release": platform.release(),
            "version": platform.version(),
            "machine": platform.machine(),
            "processor": platform.processor(),
        },
        "python": {
            "version": sys.version,
            "executable": sys.executable,
            "platform": sys.platform,
        },
        "environment": {
            "cuda_home": os.environ.get("CUDA_HOME", "Not set"),
            "cuda_path": os.environ.get("CUDA_PATH", "Not set"),
            "ld_library_path": os.environ.get("LD_LIBRARY_PATH", "Not set"),
            "path": os.environ.get("PATH", "")[:200] + "..." if len(os.environ.get("PATH", "")) > 200 else os.environ.get("PATH", ""),
        }
    }
    
    return info


def main_check_installation():
    """Main function for command-line installation checking."""
    print_installation_report()
    
    # Additional detailed system info if there are issues
    info = verify_installation()
    if info["overall_status"] in ["basic", "poor"]:
        print("\n" + "="*50)
        print("DETAILED SYSTEM INFORMATION")
        print("="*50)
        
        sys_info = get_system_info()
        print(f"\nPlatform: {sys_info['platform']['system']} {sys_info['platform']['release']}")
        print(f"Machine: {sys_info['platform']['machine']}")
        print(f"Python: {sys_info['python']['version'].split()[0]} ({sys_info['python']['executable']})")
        
        print(f"\nEnvironment:")
        print(f"  CUDA_HOME: {sys_info['environment']['cuda_home']}")
        print(f"  CUDA_PATH: {sys_info['environment']['cuda_path']}")


if __name__ == "__main__":
    main_check_installation() 